#include "ggiterm.h"
#include <string.h>
#include <stdio.h>

#define MAX_NUMERIC_SIZE 10

void ggitermEraseToEOL ();
void ggitermEraseFromBOL ();
void ggitermEraseToEOD ();
void ggitermInsertChars (unsigned int);
void ggitermEraseNChars (unsigned int);
void ggitermDeleteChars (unsigned int);
void ggitermInsertLines (unsigned int);
void ggitermDeleteLines (unsigned int);
void ggitermScrollUp (unsigned int);
void ggitermScrollDown (unsigned int);
void ggitermNextPos ();

extern unsigned short last_char;
extern int cursor_x, cursor_y, cursor_x0, cursor_y0, cell_w, cell_h;
extern int terminal_width, terminal_height;
extern int blink_mode, bold_mode, invisible_mode, reverse_mode, underline_mode, altcharset_mode;
extern int (*ggitermWriteChar) (unsigned long);


int min (int a, int b)
{
	return (b<a?b:a);
}

char ansiGetCurrentChar (char** buf, size_t* size)
{
	char current;
	
	current = (*buf)[0];
	debug (DEBUG_DETAIL,   "ansiGetCurrentChar: Character code = 0x%02X (%03d)\n", current, current);
	(*buf)++;
	(*size)--;
	return current;
}


/* This way of getting a numerical value is EVIL, because
   sscanf does not always behave consistently wrt %n across
   implementations. We should do it by hand, but before
   I need to check that we never get negative values in parameters... */
int ansiGetNumeric (char** buf, size_t* size, unsigned int* num)
{
	char szbuf[MAX_NUMERIC_SIZE+1];
	size_t szbuf_size;
	int advance, is_numeric;
	
	szbuf_size = min (*size, MAX_NUMERIC_SIZE);
	memcpy (szbuf, *buf, szbuf_size);
	szbuf[szbuf_size] = '\0';
	is_numeric = sscanf (szbuf, "%u%n", num, &advance);
	if (is_numeric)
	{
		(*buf) += advance;
		(*size) -= advance;
		debug (DEBUG_DETAIL,   "ansiGetNumeric: Numeric sequence = \"%d\"\n", *num);
	}
	return is_numeric;
}

/*------------------------------------------------------------------------------
| Cursor motion                                                                |
------------------------------------------------------------------------------*/

void do_CR ()
{
	debug (DEBUG_DETAIL, "do_CR: Special character code [CR] detected\n");
	cursor_x = cursor_x0;
}

void do_LF ()
{
	debug (DEBUG_DETAIL, "do_LF: Special character code [LF] detected\n");
	cursor_y += cell_h;
	if (cursor_y >= terminal_height * cell_h)
	{
		ggitermScrollUp (1);
		cursor_y = (terminal_height - 1) * cell_h + cursor_y0;
	}
}

void do_cub (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_cub: Escape Sequence 'cub %d' detected\n", p1);
	/* Move cursor p1 characters to the left */
	/* Should we handle the borderline case? */
	/* We shouldn't have to, since the bw cap is not in tha ANSI term definition */
	cursor_x -= (cell_w * p1);
}

void do_cub1 ()
{
	debug (DEBUG_DETAIL, "do_cub1: Escape Sequence 'cub1' or Bakspace char detected\n");
	/* Move cursor 1 character to the left */
	/* Should we handle the borderline case? */
	/* We shouldn't have to, since the bw cap is not in tha ANSI term definition */
	cursor_x -= cell_w;
}

void do_cud (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_cud: Escape Sequence 'cud %d' detected\n", p1);
	/* Move cursor down p1 lines */
	/* Should we handle the borderline case? Let's try not to */
	cursor_y += (cell_h * p1);
}

void do_cud1 ()
{
	debug (DEBUG_DETAIL, "do_cud1: Escape Sequence 'cud1' detected\n");
	/* Move cursor down 1 line */
	/* Should we handle the borderline case? Let's try not to */
	cursor_y += cell_h;
}

void do_cuf (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_cuf: Escape Sequence 'cuf %d' detected\n", p1);
	/* Move cursor p1 characters to the right */
	/* Should we handle the borderline case? Let's try not to */
	cursor_x += (cell_w * p1);
}

void do_cuf1 ()
{
	debug (DEBUG_DETAIL, "do_cuf1: Escape Sequence 'cuf1' detected\n");
	/* Move cursor 1 character to the right */
	/* Should we handle the borderline case? Let's try not to */
	cursor_x += cell_w;
}

void do_cup (unsigned int row, unsigned int column)
{
	debug (DEBUG_DETAIL, "do_cup: Escape Sequence 'cup %d %d' detected\n", row, column);
	cursor_x = cursor_x0 + (column - 1) * cell_w;
	cursor_y = cursor_y0 + (row - 1) * cell_h;
}

void do_cuu (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_cuu: Escape Sequence 'cuu %d' detected\n", p1);
	/* Move cursor up p1 lines */
	/* Should we handle the borderline case? Let's try not to */
	cursor_y -= (cell_h * p1);
}

void do_cuu1 ()
{
	debug (DEBUG_DETAIL, "do_cuu1: Escape Sequence 'cuu1' detected\n");
	/* Move cursor up 1 line */
	/* Should we handle the borderline case? Let's try not to */
	cursor_y -= cell_h;
}

void do_home ()
{
	debug (DEBUG_DETAIL, "do_home: Escape Sequence 'home' detected\n");
	cursor_x = cursor_x0;
	cursor_y = cursor_y0;
}

void do_hpa (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_hpa: Escape Sequence 'hpa %d' detected\n", p1);
	/* Horizontal position absolute */
	cursor_x = cursor_x0 + (p1-1) * cell_w;
}

void do_vpa (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_vpa: Escape Sequence 'vpa %d' detected\n", p1);
	/* Vertical position absolute */
	cursor_y = cursor_y0 + (p1-1) * cell_h;
}

/*------------------------------------------------------------------------------
| Mode handling                                                                |
------------------------------------------------------------------------------*/

void do_blink ()
{
	debug (DEBUG_DETAIL, "do_blink: Escape Sequence 'blink' or part of 'sgr' detected\n");
	debug (DEBUG_DETAIL, "do_blink: Entering blink mode\n");
	blink_mode = 1;
}

void do_bold ()
{
	debug (DEBUG_DETAIL, "do_bold: Escape Sequence 'bold' or part of 'sgr' detected\n");
	debug (DEBUG_DETAIL, "do_bold: Entering bold mode\n");
	bold_mode = 1;
}

void do_invis ()
{
	debug (DEBUG_DETAIL, "do_invis: Escape Sequence 'invis' or part of 'sgr' detected\n");
	debug (DEBUG_DETAIL, "do_invis: Entering invisible mode\n");
	invisible_mode = 1;
}

void do_rev ()
{
	debug (DEBUG_DETAIL, "do_rev: Escape Sequence 'rev|smso' or part of 'sgr' detected\n");
	debug (DEBUG_DETAIL, "do_rev: Entering reverse video mode\n");
	reverse_mode = 1;
}

void do_rmacs ()
{
	debug (DEBUG_DETAIL, "do_rmacs: Escape Sequence 'rmacs' or part of 'sgr|sgr0' detected\n");
	debug (DEBUG_DETAIL, "do_rmacs: Leaving altcharset mode\n");
	altcharset_mode = 0;
}

void do_rmso ()
{
	debug (DEBUG_DETAIL, "do_rmso: Escape Sequence 'rmso|rmul' or part of 'sgr|sgr0' detected\n");
	debug (DEBUG_DETAIL, "do_rmso: Leaving all modes (except altcharset)\n");
	bold_mode = 0;
	blink_mode = 0;
	underline_mode = 0;
	reverse_mode = 0;
	invisible_mode = 0;
}

void do_smacs ()
{
	debug (DEBUG_DETAIL, "do_smacs: Escape Sequence 'smacs' or part of 'sgr' detected\n");
	debug (DEBUG_DETAIL, "do_smacs: Entering altcharset mode\n");
	altcharset_mode = 1;
}

void do_smul ()
{
	debug (DEBUG_DETAIL, "do_smul: Escape Sequence 'smul' or part of 'sgr' detected\n");
	debug (DEBUG_DETAIL, "do_smul: Entering underline mode\n");
	underline_mode = 1;
}

/*------------------------------------------------------------------------------
| Area clear                                                                   |
------------------------------------------------------------------------------*/

void do_ed ()
{
	debug (DEBUG_DETAIL, "do_ed: Escape Sequence 'ed' detected\n");
	ggitermEraseToEOD ();
}

void do_el ()
{
	debug (DEBUG_DETAIL, "do_el: Escape Sequence 'el' detected\n");
	ggitermEraseToEOL ();
}

void do_el1 ()
{
	debug (DEBUG_DETAIL, "do_el1: Escape Sequence 'el1' detected\n");
	ggitermEraseFromBOL ();
}

/*------------------------------------------------------------------------------
| Insert/Delete characters                                                     |
------------------------------------------------------------------------------*/

void do_ich (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_ich: Escape Sequence 'ich %d' detected\n", p1);
	/* Insert p1 characters at current position */
	ggitermInsertChars (p1);
}

void do_dch (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_dch: Escape Sequence 'dch %d' detected\n", p1);
	/* Delete p1 characters */
	ggitermDeleteChars (p1);
}

void do_dch1 ()
{
	debug (DEBUG_DETAIL, "do_dch1: Escape Sequence 'dch1' detected\n");
	/* Delete 1 character */
	ggitermDeleteChars (1);
}

void do_ech (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_ech: Escape Sequence 'ech %d' detected\n", p1);
	/* Erase p1 characters, starting at current position */
	ggitermEraseNChars (p1);
}

/*------------------------------------------------------------------------------
| Insert/Delete lines                                                          |
------------------------------------------------------------------------------*/

void do_dl (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_dl: Escape Sequence 'dl %d' detected\n", p1);
	/* Delete p1 lines at current position */
	ggitermDeleteLines (p1);
}

void do_dl1 ()
{
	debug (DEBUG_DETAIL, "do_dl1: Escape Sequence 'dl1' detected\n");
	/* Delete 1 line at current position */
	ggitermDeleteLines (1);
}

void do_il (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_il: Escape Sequence 'il %d' detected\n", p1);
	/* Insert p1 lines at current position */
	ggitermInsertLines (p1);
}

void do_il1 ()
{
	debug (DEBUG_DETAIL, "do_il1: Escape Sequence 'il1' detected\n");
	/* Insert 1 line at current position */
	ggitermInsertLines (1);
}

/*------------------------------------------------------------------------------
| Scrolling                                                                    |
------------------------------------------------------------------------------*/

void do_ind ()
{
	debug (DEBUG_DETAIL, "do_ind: Escape Sequence 'ind' detected\n");
	/* Scroll forward 1 line */
	/* Never ued, because ansi says ind=ctrl-J (CR+LF) and is only called
	   when we are on the last line, which amounts to scroll up 1 line */
	ggitermScrollUp (1);
}

void do_indn (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_indn: Escape Sequence 'indn %d' detected\n", p1);
	/* Scroll forward p1 lines */
	ggitermScrollUp (p1);
}

void do_rin (unsigned int p1)
{
	debug (DEBUG_DETAIL, "do_rin: Escape Sequence 'rin %d' detected\n", p1);
	/* Scroll back p1 lines */
	ggitermScrollDown (p1);
}

/*------------------------------------------------------------------------------
| Tabs                                                                         |
------------------------------------------------------------------------------*/

void do_cbt ()
{
	debug (DEBUG_DETAIL, "do_cbt: Escape Sequence 'cbt' detected\n");
}

void do_ht ()
{
	debug (DEBUG_DETAIL, "do_ht: Escape Sequence 'ht' detected\n");
}

void do_hts ()
{
	debug (DEBUG_DETAIL, "do_hts: Escape Sequence 'hts' detected\n");
}

void do_tbc ()
{
	debug (DEBUG_DETAIL, "do_tbc: Escape Sequence 'tbc' detected\n");
}

/*------------------------------------------------------------------------------
| Others                                                                       |
------------------------------------------------------------------------------*/

void do_s0ds ()
{
	debug (DEBUG_DETAIL, "do_s0ds: Escape Sequence 's0ds' detected\n");
	/* Shift to codeset 0 (ASCII, EUC set 0) */
	/* ASCII is the default for us, and there is no other codeset */
	/* This is a NOOP here */
}

void do_s1ds ()
{
	debug (DEBUG_DETAIL, "do_s1ds: Escape Sequence 's1ds' detected\n");
	/* Shift to codeset 1 */
	/* This is a NOOP here */
}

void do_s2ds ()
{
	debug (DEBUG_DETAIL, "do_s2ds: Escape Sequence 's2ds' detected\n");
	/* Shift to codeset 2 */
	/* This is a NOOP here */
}

void do_s3ds ()
{
	debug (DEBUG_DETAIL, "do_s3ds: Escape Sequence 's3ds' detected\n");
	/* Shift to codeset 3 */
	/* This is a NOOP here */
}

void do_mc4 ()
{
	debug (DEBUG_DETAIL, "do_mc4: Escape Sequence 'mc4' detected\n");
	/* Turn off printer */
	/* This is a NOOP here */
}

void do_mc5 ()
{
	debug (DEBUG_DETAIL, "do_mc5: Escape Sequence 'mc5' detected\n");
	/* Turn on printer */
	/* This is a NOOP here */
}

void do_BELL ()
{
	debug (DEBUG_DETAIL, "do_BELL: Special character code [BELL] detected\n");
	/* This should be a visual bell */
}

void do_rep (unsigned int count)
{
	debug (DEBUG_DETAIL, "do_rep: Escape Sequence 'rep %d' detected\n", count);
	/* Repeat last character count times */
	while (count)
	{
		ggitermWriteChar (last_char);
		ggitermNextPos ();
		count--;
	}
}


int
HandleEscapeSequences (char **buf, size_t *bufsize)
{
	size_t size;
	char *escape;
	unsigned int param[9];
	int is_numeric, num_param, i;
	char current_char;

	escape = *buf;
	size = *bufsize;
	current_char = ansiGetCurrentChar (&escape, &size);
	switch (current_char)
	{
	case 0x0D:
		/* Carriage return (13) */
		do_CR ();
		break;
	case 0x0A:
		/* Line feed (10) */
		do_LF ();
		break;
	case 0x07:
		/* Bell (7) */
		do_BELL ();
		break;
	case 0x08:
		/* Backspace */
		do_cub1 ();
		break;
	case 27:		/*  Escape character */
		if (!size) return 2;
		current_char = ansiGetCurrentChar (&escape, &size);
		switch (current_char)
		{
		case '+':
			if (!size) return 2;
			current_char = ansiGetCurrentChar (&escape, &size);
			switch (current_char)
			{
				case 'B':
					do_s3ds ();
					break;
				default:
					return 1;
			}
			break;
		case '*':
			if (!size) return 2;
			current_char = ansiGetCurrentChar (&escape, &size);
			switch (current_char)
			{
				case 'B':
					do_s2ds ();
					break;
				default:
					return 1;
			}
			break;
		case ')':
			if (!size) return 2;
			current_char = ansiGetCurrentChar (&escape, &size);
			switch (current_char)
			{
				case 'B':
					do_s1ds ();
					break;
				default:
					return 1;
			}
			break;
		case '(':
			if (!size) return 2;
			current_char = ansiGetCurrentChar (&escape, &size);
			switch (current_char)
			{
				case 'B':
					do_s0ds ();
					break;
				default:
					return 1;
			}
			break;
		case '[':
			if (!size) return 2;
			is_numeric = ansiGetNumeric (&escape, &size, &(param[0]));
			if (!size) return 2;
			current_char = ansiGetCurrentChar (&escape, &size);
			switch (current_char)
			{
				case 'A':
					if (is_numeric) do_cuu (param[0]);
					else do_cuu1 ();
					break;
				case 'B':
					if (is_numeric) do_cud (param[0]);
					else do_cud1 ();
					break;
				case 'C':
					if (is_numeric) do_cuf (param[0]);
					else do_cuf1 ();
					break;
				case 'D':
					if (is_numeric) do_cub (param[0]);
					else do_cub1 ();
					break;
				case 'G':
					if (is_numeric) do_hpa (param[0]);
					else return 1;
					break;
				case 'H':
					if (is_numeric) return 1;
					else do_home ();
					break;
				case 'I':
					if (is_numeric) return 1;
					else do_ht ();
					break;
				case 'J':
					if (is_numeric) return 1;
					else do_ed ();
					break;
				case 'K':
					if (is_numeric)
					{
						if (param[0]==1) do_el1 ();
						else return 1;
					}
					else do_el ();
					break;
				case 'L':
					if (is_numeric) do_il (param[0]);
					else do_il1 ();
					break;
				case 'M':
					if (is_numeric) do_dl (param[0]);
					else do_dl1 ();
					break;
				case 'P':
					if (is_numeric) do_dch (param[0]);
					else do_dch1 ();
					break;
				case 'S':
					if (is_numeric) do_indn (param[0]);
					else do_LF();
					break;
				case 'T':
					if (is_numeric) do_rin (param[0]);
					else return 1;
					break;
				case 'X':
					if (is_numeric) do_ech (param[0]);
					else return 1;
					break;
				case 'Z':
					if (is_numeric) return 1;
					else do_cbt ();
					break;
				case 'b':
					if (is_numeric) do_rep (param[0]);
					else return 1;
					break;
				case 'd':
					if (is_numeric) do_vpa (param[0]);
					else return 1;
					break;
				case 'g':
					if (is_numeric && param[0]==2) do_tbc ();
					else return 1;
					break;
				case 'i':
					if (is_numeric)
					{
						switch (param[0])
						{
							case 4:
								do_mc4();
								break;
							case 5:
								do_mc5();
								break;
							default:
								return 1;
						}
					}
					else return 1;
					break;
				case '@':
					if (is_numeric) do_ich (param[0]);
					else return 1;
					break;
				case 'm':
				case ';':
					if (!is_numeric) param[0] = 0;
					num_param = 1;
					while (current_char == ';' && num_param < 9)
					{
						if (!size) return 2;
						is_numeric = ansiGetNumeric (&escape, &size, &(param[num_param]));
						if (!is_numeric) return 1;
						if (!size) return 2;
						current_char = ansiGetCurrentChar (&escape, &size);
						num_param++;
					}
					switch (current_char)
					{
						case 'H':
							if (num_param == 2) do_cup (param[0], param[1]);
							else return 1;
							break;
						case 'm':
							for (i = 0; i < num_param; i++)
							{
								switch (param[i])
								{
									case 0:
										do_rmso ();
										break;
									case 1:
										do_bold ();
										break;
									case 4:
										do_smul ();
										break;
									case 5:
										do_blink ();
										break;
									case 7:
										do_rev ();
										break;
									case 8:
										do_invis ();
										break;
									case 10:
										do_rmacs ();
										break;
									case 12:
										do_smacs ();
										break;
									default:
										return 1;
								}
							}
							break;
						default:
							return 1;
					}
					break;
				default:
					return 1;
			}
			break;
		case 'H':
			do_hts ();
			break;
		default:
			return 1;
		}
		break;
	default:
		return 1;
	}
	
	*buf = escape;
	*bufsize = size;
	
	return 0;
}
