
#include "ggiterm.h"

#if HAVE_UNISTD_H
# include <unistd.h>
#else
#				/* We are in trouble.  */
#endif

#include <stdio.h>
#include <stdlib.h>


/* error handling */
#include <errno.h>

/* signals */
#include <signal.h>

/* for forkpty */
#ifdef HAVE_FORKPTY
/* forkpty is in libc */
#include <util.h>
#else
/* forkpty is in libutil */
#include <pty.h>
#endif /* HAVE_FORKPTY */

/* for strcmp */
#include <string.h>

/* for sleep, select */
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>


/* must be defined because even if FT support is not compiled in,
   FreetypeInit() is called with fontsize as argument */
#define DEFAULT_FONTSIZE 14

/* External functions and variables */
int FreetypeInit (char *, int, int (**) (unsigned long));
void FreetypeExit ();
#ifdef DEBUG
void FreetypeShowCharmap ();
#endif /* DEBUG */

int HandleEscapeSequences (char **, size_t *);

void ggitermNextPos ();
int GGIWriteChar (unsigned long);
int GGIInit_graphicsmode (void);
int GGIInit_textmode (void);
void HandleGGIEvents (char*, int*);
void GGIExit (void);
void ggitermBlinkCursor (int);
#ifdef DEBUG
void ShowGrid (void);
#endif /* DEBUG */

extern char* encoding_from;
int IconvInit (int (**) (unsigned short*, char**, size_t*));
void IconvExit ();

extern int vis_w, vis_h;
extern int cursor_x, cursor_y, cursor_x0, cursor_y0, cell_w, cell_h;

/* Global variables */
int terminal_width, terminal_height;
int quit = 0;
#ifdef DEBUG
int debuglevel = 0;
#endif /* DEBUG */
unsigned short last_char;

int (*ggitermWriteChar) (unsigned long charcode);
int (*ggitermGetNextChar) (unsigned short *charcode, char **inputbuf,
			   size_t * inputsize);


int
DefaultGetNextChar (unsigned short *charcode, char **inputbuf,
		    size_t * inputsize)
{
	*charcode = (unsigned char) (*inputbuf)[0];
	(*inputsize)--;
	(*inputbuf)++;
	return 0;
}


void
ggitermUsage (void)
{
	fprintf (stderr,
		 "\n"
		 "ggiterm v%s\n"
		 "\n"
		 "Usage: ggiterm [OPTIONS...]\n"
		 "Where option is one or more of:\n"
		 "	--help\n", VERSION);
#ifdef DEBUG
	fprintf (stderr,
		"	--debuglevel n           (%d:none, %d:info, %d:detailed, %d:total, default:%d)\n",
		DEBUG_NONE, DEBUG_INFO, DEBUG_DETAIL, DEBUG_TOTAL, DEBUG_NONE);
#endif /* DEBUG */
#ifdef HAVE_FREETYPE
	fprintf (stderr, "	--font <font file>\n");
	fprintf (stderr, "	--font-size n            (default: %d)\n",
		 DEFAULT_FONTSIZE);
#endif /* HAVE_FREETYPE */
#ifdef HAVE_ICONV
	fprintf (stderr,
		 "	--encoding <latin1|utf8> (default: %s)\n", encoding_from);
#endif /* HAVE_ICONV */
#ifdef DEBUG
	fprintf (stderr, "\nDebug support:    yes\n");
#else
	fprintf (stderr, "\nDebug support:    no\n");
#endif /* DEBUG */
#ifdef HAVE_FREETYPE
	fprintf (stderr, "Freetype support: yes\n");
#else
	fprintf (stderr, "Freetype support: no\n");
#endif /* HAVE_FREETYPE */
#ifdef HAVE_ICONV
	fprintf (stderr, "Iconv support:    yes\n\n");
#else
	fprintf (stderr, "Iconv support:    no\n\n");
#endif /* HAVE_ICONV */
}

RETSIGTYPE
child_died ()
{
	quit = 2;
}


void
HandleIncomingChars (char *buf, size_t * size)
{
	int err;
	size_t inputsize;
	char *inputbuf;
	unsigned short charcode;
	/*char dummy[1025];*/

	inputbuf = buf;
	inputsize = *size;
	/*strncpy (dummy, inputbuf, inputsize);
	dummy[inputsize] = '\0';
	debug (DEBUG_INFO, "HandleIncomingChars: Buffer contents: [\n%s\n]\n",
	       dummy);*/
	do
	{
		do
		{
			err = HandleEscapeSequences (&inputbuf, &inputsize);
			debug (DEBUG_DETAIL, "HandleIncomingChars: HandleEscapeSequences returned %d\n", err);
		}
		while (!err && inputsize > 0);
		if (err == 1 && inputsize > 0)
		{
			err = ggitermGetNextChar (&charcode, &inputbuf, &inputsize);
			if (!err)
			{
				debug (DEBUG_DETAIL,
					       "HandleIncomingChars: Character code = 0x%04X (%05d)\t\"%c\"\n",
					       charcode, charcode, charcode);
				ggitermWriteChar (charcode);
				last_char = charcode;
				ggitermNextPos ();
			}
		}
	}
	while (inputsize > 0 && err != 2);
	memmove (buf, inputbuf, inputsize);
	*size = inputsize;
}


int
main (int argc, char **argv, char **envp)
{
	int err, using_ft=0, using_iconv=0;
	char *font_file = NULL;
	/* must be defined because even if FT support is not compiled in,
	   FreetypeInit() is called with fontsize as argument */
	int fontsize = DEFAULT_FONTSIZE;
	pid_t shell_pid;
	char buf[1024];
	size_t bufsize = 0;
	int master_fd;
	struct timeval tv;
	fd_set rfds, wfds;
	char inputbuffer[10];
	int inputbuffer_size = 0;
	char *shell = "/bin/sh";
	struct winsize ws;


	argv++;
	argc--;
	while (argc > 0)
	{
		if (!strcmp (argv[0], "--help"))
		{
			ggitermUsage ();
			return 0;
		}
#ifdef HAVE_FREETYPE
		if (!strcmp (argv[0], "--font"))
		{
			font_file = argv[1];
			argc -= 2;
			argv += 2;
			continue;
		}
		if (!strcmp (argv[0], "--font-size"))
		{
			fontsize = strtol (argv[1], NULL, 10);
			if (errno == ERANGE)
			{
				perror ("ggiterm: strtol");
				return 1;
			}
			argc -= 2;
			argv += 2;
			continue;
		}
#endif /* HAVE_FREETYPE */
#ifdef HAVE_ICONV
		if (!strcmp (argv[0], "--encoding"))
		{
			encoding_from = argv[1];
			argc -= 2;
			argv += 2;
			continue;
		}
#endif /* HAVE_ICONV */
#ifdef DEBUG
		if (!strcmp (argv[0], "--debuglevel"))
		{
			debuglevel = strtol (argv[1], NULL, 10);
			if (errno == ERANGE)
			{
				perror ("ggiterm: strtol");
				return 1;
			}
			argc -= 2;
			argv += 2;
			continue;
		}
#endif /* DEBUG */
		fprintf (stderr, "main: Unknown option %s\n", argv[0]);
		return 1;
	}

	debug (DEBUG_INFO, "main: Setting up default handlers:\n");
	ggitermGetNextChar = &DefaultGetNextChar;
	debug (DEBUG_INFO, "main:        ggitermGetNextChar -> DefaultGetNextChar\n");
	ggitermWriteChar = &GGIWriteChar;
	debug (DEBUG_INFO, "main:        ggitermWriteChar   -> GGIWriteChar\n");
	
	err = GGIInit_graphicsmode ();
	if (!err)
	{
		debug (DEBUG_INFO, "main: Using GGI in graphics mode.\n");
		if (font_file != NULL)
		{
			if (!FreetypeInit (font_file, fontsize, &ggitermWriteChar))
			{
				debug (DEBUG_INFO, "main: Using Freetype rendering.\n");
				using_ft = 1;
				if (!IconvInit (&ggitermGetNextChar))
				{
					debug (DEBUG_INFO,  "main: Using Iconv character translation.\n");
					using_iconv = 1;
				}
				else
				{
					fprintf (stderr, "main: Cannot use Iconv character translation.\n");
				}
			}
			else
			{
				fprintf (stderr, "main: Cannot use Freetype rendering.\n");
				fprintf (stderr, "main: Staying in native GGI font rendering (ASCII encoding only).\n");
			}
		}
		else
		{
			debug (DEBUG_INFO, "main: No font file provided.\n");
			debug (DEBUG_INFO, "main: Using native GGI font rendering (ASCII encoding only).\n");
		}	
	}
	else
	{
		fprintf (stderr, "main: Cannot use GGI in graphics mode. Trying text mode...\n");
		err = GGIInit_textmode ();
		if (!err)
		{
		}
		else
		{
			fprintf (stderr,
				 "main: Cannot complete GGI setup. Aborting.\n");
			GGIExit ();
			return 1;
		}
	}
	terminal_width = vis_w / cell_w;
	terminal_height = vis_h / cell_h;
	debug (DEBUG_INFO, "main: Terminal is %d x %d characters\n",
	       terminal_width, terminal_height);
	debug (DEBUG_INFO, "main: Characters are %d x %d pixels\n", cell_w,
	       cell_h);

#ifdef DEBUG
	ShowGrid ();
	FreetypeShowCharmap ();
#endif /* DEBUG */


	if (getenv ("SHELL") != NULL)
	{
		shell = getenv ("SHELL");
	}
	signal (SIGCHLD, child_died);

	setenv ("TERM", "ansi-mono", 1);
	setenv ("PS1", "ggiterm> ", 1);

	ws.ws_col = terminal_width;
	ws.ws_row = terminal_height;
	ws.ws_xpixel = vis_w;
	ws.ws_ypixel = vis_h;

	/* in the following, buf should be set to NULL because of buffer overflow
	 * risk, but is left so for debugging purposes */
	shell_pid = forkpty (&master_fd, buf, NULL, &ws);
	if (shell_pid < 0)
	{
		perror ("ggiterm: Couldn't fork. Exiting.\n");
		exit (1);
	}
	if (shell_pid == 0)
	{
		char *argv[] = { shell, NULL };

		debug (DEBUG_INFO,
			 "\nThis is the forked process (PID %d, child of PID %d)\n\n",
			 getpid (), getppid ());
		debug (DEBUG_INFO,
			 "Sample UTF-8 output: Caractères français\n");
		debug (DEBUG_INFO,
			 "Sample LATIN1 output: Caractres franais\n");
		debug (DEBUG_INFO, "Now exec-ing %s\n", shell);
		execve (argv[0], argv, envp);
		exit (1);
	}
	debug (DEBUG_INFO, "main: Child forked (PID %d), pseudo tty is %s\n",
	       shell_pid, buf);
	debug (DEBUG_INFO, "main: Waiting for its death...\n");

	while (!quit)
	{
		ggitermBlinkCursor (0);
		
		/* do we have pending GGI events?
		 * (non blocking)
		 */
		HandleGGIEvents (inputbuffer, &inputbuffer_size);

		tv.tv_sec = 0;
		tv.tv_usec = 20000;
		FD_ZERO (&rfds);
		FD_ZERO (&wfds);
		FD_SET (master_fd, &rfds);
		if (inputbuffer_size)
			FD_SET (master_fd, &wfds);

		/*do
		{*/
			err = select (master_fd + 1, &rfds, &wfds, NULL, &tv);
		/*}
		while (err < 0 && errno == EINTR && !quit);*/
		if (err < 0 && errno!=EINTR)
		{
			perror ("ggiterm: select");
		}
		if (err < 0 && errno==EINTR)
		{
			continue;
		}
		if (quit)
			break;
		if (err > 0)
			debug (DEBUG_DETAIL,
			       "main: %d file descriptors updated\n", err);
		

		if (FD_ISSET (master_fd, &wfds))
		{
			debug (DEBUG_DETAIL,
			       "main: master_fd is ready for writing\n");
			/* if we have anything pending, then write it */
			if (inputbuffer_size)
			{
				write (master_fd, inputbuffer,
				       inputbuffer_size);
				debug (DEBUG_DETAIL,
				       "main: Written %d bytes to master_fd\n",
				       inputbuffer_size);
				inputbuffer_size = 0;
			}
		}

		if (FD_ISSET (master_fd, &rfds))
		{
			debug (DEBUG_DETAIL,
			       "main: master_fd is set\nmain: Reading:\n");
			debug (DEBUG_DETAIL, "main: Buffer index is %d\n",
			       bufsize);
			do
			{
				err = read (master_fd, buf + bufsize,
					    1024 - bufsize);
			}
			while (err < 0 && errno != EINTR && !quit);
			if (err < 0)
			{
				perror ("ggiterm: read");
			}
			if (err > 0)
			{
				debug (DEBUG_DETAIL,
				       "main: Read %d bytes from master_fd\n",
				       err);
				bufsize += err;
				debug (DEBUG_DETAIL,
				       "main: Buffer index is now %d\n",
				       bufsize);
				ggitermBlinkCursor (1);
				HandleIncomingChars (buf, &bufsize);
				ggitermBlinkCursor (2);
			}
			if (err == 0)
			{
				quit = 1;
			}
			debug (DEBUG_DETAIL, "\nmain: Finished reading\n");
		}
	}


	switch (quit)
	{
	case 1:
		debug (DEBUG_INFO, "main: Exiting: EOF condition\n");
		break;
	case 2:
		debug (DEBUG_INFO, "main: Exiting: Child process has died\n");
		break;
	case 3:
		debug (DEBUG_INFO, "main: Exiting: Pipe error\n");
		break;
	}
	ggitermBlinkCursor (1);
	if (using_ft)
	{
		FreetypeExit ();
		if (using_iconv) IconvExit ();
	}
	GGIExit ();
	
	return (0);
}
